#include "collision_manager_alg_node.h"

CollisionManagerAlgNode::CollisionManagerAlgNode(void) :
  algorithm_base::IriBaseAlgorithm<CollisionManagerAlgorithm>()
  ,tf2_listener(tf2_buffer)
{
  //init class attributes if necessary
  if(!this->private_node_handle_.getParam("rate", this->config_.rate))
  {
    ROS_WARN("CollisionManagerAlgNode::CollisionManagerAlgNode: param 'rate' not found");
  }
  else
    this->setRate(this->config_.rate);

  if(!this->private_node_handle_.getParam("base_link_frame", this->config_.base_link_frame))
  {
    ROS_WARN("AdcLandmarksSlamSolverAlgNode::AdcLandmarksSlamSolverAlgNode: param 'base_link' not found. Setting it to base_link.");
    this->config_.base_link_frame = "base_link";
  }

  if(!this->private_node_handle_.getParam("fixed_frame", this->config_.fixed_frame))
  {
    ROS_WARN("AdcLandmarksSlamSolverAlgNode::AdcLandmarksSlamSolverAlgNode: param 'base_link' not found. Setting it to map.");
    this->config_.fixed_frame = "map";
  }

  this->new_imu_input_ = false;
  this->imu_frame_ = "";
  this->front_laser_frame_ = "";
  this->rear_laser_frame_ = "";
  this->tf_base_link_imu_loaded_ = false;
  this->in_collision_ = false;
  this->low_acc_count_ = 0;
  this->low_ang_vel_count_ = 0;
  this->is_front_icp_input_data_ok_ = false;
  this->is_rear_icp_input_data_ok_ = false;
  this->tf_base_link_front_laser_loaded_ = false;
  this->tf_base_link_rear_laser_loaded_ = false;
  this->angular_vel_ = 0.0;
  this->odom_input_is_ok_ = false;

  this->debug_rear_icp_pose_msg_.header.frame_id = this->config_.fixed_frame;
  this->debug_rear_icp_pose_msg_.pose.pose.position.z = 0.0;
  for (unsigned int i = 0; i < 36; i++)
    this->debug_rear_icp_pose_msg_.pose.covariance[i] = 0.0;

  this->debug_front_icp_pose_msg_.header.frame_id = this->config_.fixed_frame;
  this->debug_front_icp_pose_msg_.pose.pose.position.z = 0.0;
  for (unsigned int i = 0; i < 36; i++)
    this->debug_front_icp_pose_msg_.pose.covariance[i] = 0.0;

  this->collision_msg_.header.frame_id = "iri_collision_manager";

  // [init publishers]
  this->collisions_publisher_ = this->private_node_handle_.advertise<iri_collision_manager::collision>("collisions", 1);
  this->debug_rear_after_icp_scan_publisher_ = this->private_node_handle_.advertise<sensor_msgs::LaserScan>("debug_rear_after_icp_scan", 1);
  this->debug_front_after_icp_scan_publisher_ = this->private_node_handle_.advertise<sensor_msgs::LaserScan>("debug_front_after_icp_scan", 1);
  this->debug_rear_icp_pose_publisher_ = this->private_node_handle_.advertise<geometry_msgs::PoseWithCovarianceStamped>("debug_rear_icp_pose", 1);
  this->debug_front_icp_pose_publisher_ = this->private_node_handle_.advertise<geometry_msgs::PoseWithCovarianceStamped>("debug_front_icp_pose", 1);
  this->debug_rear_icp_last_scan_publisher_ = this->private_node_handle_.advertise<sensor_msgs::LaserScan>("debug_rear_icp_last_scan", 1);
  this->debug_rear_icp_ref_scan_publisher_ = this->private_node_handle_.advertise<sensor_msgs::LaserScan>("debug_rear_icp_ref_scan", 1);
  this->debug_front_icp_last_scan_publisher_ = this->private_node_handle_.advertise<sensor_msgs::LaserScan>("debug_front_icp_last_scan", 1);
  this->debug_front_icp_ref_scan_publisher_ = this->private_node_handle_.advertise<sensor_msgs::LaserScan>("debug_front_icp_ref_scan", 1);
  
  // [init subscribers]
  this->odom_subscriber_ = this->private_node_handle_.subscribe("odom", 1, &CollisionManagerAlgNode::odom_callback, this);
  pthread_mutex_init(&this->odom_mutex_,NULL);

  this->rear_laser_scan_subscriber_ = this->private_node_handle_.subscribe("rear_laser_scan", 1, &CollisionManagerAlgNode::rear_laser_scan_callback, this);
  pthread_mutex_init(&this->rear_laser_scan_mutex_,NULL);

  this->front_laser_scan_subscriber_ = this->private_node_handle_.subscribe("front_laser_scan", 1, &CollisionManagerAlgNode::front_laser_scan_callback, this);
  pthread_mutex_init(&this->front_laser_scan_mutex_,NULL);

  this->imu_subscriber_ = this->private_node_handle_.subscribe("imu", 1, &CollisionManagerAlgNode::imu_callback, this);
  pthread_mutex_init(&this->imu_mutex_,NULL);

  
  // [init services]
  
  // [init clients]
  icp_client_ = this->private_node_handle_.serviceClient<iri_laser_scan_icp::icp>("icp");

  
  // [init action servers]
  
  // [init action clients]
}

CollisionManagerAlgNode::~CollisionManagerAlgNode(void)
{
  // [free dynamic memory]
  pthread_mutex_destroy(&this->odom_mutex_);
  pthread_mutex_destroy(&this->rear_laser_scan_mutex_);
  pthread_mutex_destroy(&this->front_laser_scan_mutex_);
  pthread_mutex_destroy(&this->imu_mutex_);
}

void CollisionManagerAlgNode::mainNodeThread(void)
{
  //lock access to algorithm if necessary
  // this->alg_.lock();
  // ROS_DEBUG("CollisionManagerAlgNode::mainNodeThread");
  this->imu_mutex_enter();
  this->rear_laser_scan_mutex_enter();
  this->front_laser_scan_mutex_enter();
  this->odom_mutex_enter();
  Eigen::Vector3d blink_acc, blink_ang_vel, blink_pose, new_laser_pose, icp_odometry, icp_odometry_diff, diff_wheel_icp_odom;
  Eigen::Matrix3d blink_cov, covariances, icp_odom_cov;
  bool icp_succedded, icp_odom_calculated, icp_blink_pose_calculated, icp_is_collision;

  if (this->new_imu_input_)
  {
    if (get_base_link_imu_data(blink_acc, blink_ang_vel))//Transform data to fixed frame coordinates
    {
      double acc_norm_2d = std::sqrt(std::pow(blink_acc(0),2) + std::pow(blink_acc(1),2));
      if (this->in_collision_)
      {
        double max_acc_norm_2d = std::sqrt(std::pow(this->max_acc_(0),2) + std::pow(this->max_acc_(1),2));
        if (check_end_of_collision(acc_norm_2d, std::fabs(blink_ang_vel(2))))
        {
          double collision_angle = std::atan2(this->max_acc_(1), this->max_acc_(0));
          while (collision_angle >= M_PI)
            collision_angle -= 2*M_PI;
          while (collision_angle < -M_PI)
            collision_angle += 2*M_PI;
          ROS_INFO_STREAM("CollisionManagerAlgNode::mainNodeThread -> collision end detected.");
          ROS_INFO_STREAM("CollisionManagerAlgNode::mainNodeThread -> collision detected: acc = " << max_acc_norm_2d << "; angle = " << collision_angle);

          //update collision msg data
          this->collision_msg_.acc_peak = max_acc_norm_2d;
          this->collision_msg_.acc_angle = collision_angle;

          //update rot_angle_estimated
          this->front_icp_srv_.request.rot_angle_estimation += blink_ang_vel(2)*(this->imu_stamp_ - this->last_imu_t_).toSec();

          //rotate angle when fixed tf

          this->rear_icp_srv_.request.rot_angle_estimation = this->front_icp_srv_.request.rot_angle_estimation;

          // ROS_INFO_STREAM("CollisionManagerAlgNode::mainNodeThread -> collision detected: angle estimated = " << this->front_icp_srv_.request.rot_angle_estimation);

          this->odom_input_is_ok_ = (this->odom_input_is_ok_ && get_last_odom());
          // Call the service
          if (!(this->config_.front_laser_en || this->config_.rear_laser_en))
            ROS_WARN("CollisionManagerAlgNode::mainNodeThread -> Collision end detected but no laser is enabled.");
          else
          {
            if (this->is_front_icp_input_data_ok_ && set_last_ranges(true))
            {
              if (icp_client_.call(front_icp_srv_))
              {
                // ROS_INFO("CollisionManagerAlgNode::mainNodeThread -> Front icp_client_ received a response from service server");
                // ROS_INFO_STREAM("   valid " << (int)front_icp_srv_.response.valid);
                // ROS_INFO_STREAM("   new_laser_pose " << front_icp_srv_.response.new_laser_pose);
                // ROS_INFO_STREAM("   covariance " << front_icp_srv_.response.covariance[0] << ", " << front_icp_srv_.response.covariance[1] << ", " << front_icp_srv_.response.covariance[2] << ", " << front_icp_srv_.response.covariance[3] << ", " << front_icp_srv_.response.covariance[4] << ", " << front_icp_srv_.response.covariance[5] << ", " << front_icp_srv_.response.covariance[6] << ", " << front_icp_srv_.response.covariance[7] << ", " << front_icp_srv_.response.covariance[8]);
                // ROS_INFO_STREAM("   nvalid " << front_icp_srv_.response.nvalid << "; min " << this->config_.icp_min_points);
                // ROS_INFO_STREAM("   error " << front_icp_srv_.response.error);
                // ROS_INFO_STREAM("   error/points " << front_icp_srv_.response.error/front_icp_srv_.response.nvalid << "; max " << this->config_.icp_max_err_by_points);

                icp_succedded = true;

                //update collison msg data.
                this->collision_msg_.front_icp_err = front_icp_srv_.response.error/front_icp_srv_.response.nvalid;
                this->collision_msg_.front_icp_points = front_icp_srv_.response.nvalid;

                //calculate new base_link pose in fixed frame coordenates
                new_laser_pose << front_icp_srv_.response.new_laser_pose.x, front_icp_srv_.response.new_laser_pose.y, front_icp_srv_.response.new_laser_pose.theta;
                covariances << front_icp_srv_.response.covariance[0], front_icp_srv_.response.covariance[1], front_icp_srv_.response.covariance[2],
                                front_icp_srv_.response.covariance[3], front_icp_srv_.response.covariance[4], front_icp_srv_.response.covariance[5],
                                front_icp_srv_.response.covariance[6], front_icp_srv_.response.covariance[7], front_icp_srv_.response.covariance[8];

                icp_blink_pose_calculated = get_base_link_new_pose_from_icp(new_laser_pose, covariances, blink_pose, blink_cov, true);

                //publish new laser tf to visualize the laser points.
                if (this->config_.debug)
                {
                  this->transform_msg.header.frame_id = this->front_laser_frame_;
                  this->transform_msg.child_frame_id = "new_front_laser";
                  this->transform_msg.header.stamp = this->imu_stamp_;
                  this->transform_msg.transform.translation.x = front_icp_srv_.response.new_laser_pose.x;
                  this->transform_msg.transform.translation.y = front_icp_srv_.response.new_laser_pose.y;
                  this->transform_msg.transform.translation.z = 0.0;
                  tf2::Quaternion q;
                  q.setRPY(0.0, 0.0, front_icp_srv_.response.new_laser_pose.theta);
                  this->transform_msg.transform.rotation = tf2::toMsg(q);
                  // this->transform_msg.transform.rotation.x = q.getRotation().x();
                  // this->transform_msg.transform.rotation.y = q.getRotation().y();
                  // this->transform_msg.transform.rotation.z = q.getRotation().z();
                  // this->transform_msg.transform.rotation.w = q.getRotation().w();
                  this->tf2_broadcaster.sendTransform(this->transform_msg);
                }

                if (front_icp_srv_.response.nvalid > this->config_.icp_min_points && front_icp_srv_.response.error/front_icp_srv_.response.nvalid < this->config_.icp_max_err_by_points)
                {
                  if (get_icp_odometry(new_laser_pose, covariances, icp_odometry, icp_odom_cov, icp_odometry_diff, true))
                  {
                    icp_odom_calculated = true;
                    ROS_INFO_STREAM("Odom before collision " << this->odom_before_collision_);
                    ROS_INFO_STREAM("Odom after collision " << this->odom_after_collision_);
                    ROS_INFO_STREAM("New laser pose " << new_laser_pose);
                    ROS_INFO_STREAM("ICP odom " << icp_odometry);
                    ROS_INFO_STREAM("ICP odom diff " << icp_odometry_diff);
                    icp_is_collision = check_odometry_diff(icp_odometry_diff, diff_wheel_icp_odom);

                    // Update collision msg
                    this->collision_msg_.front_odom_diff.x = diff_wheel_icp_odom(0);
                    this->collision_msg_.front_odom_diff.y = diff_wheel_icp_odom(1);
                    this->collision_msg_.front_odom_diff.theta = diff_wheel_icp_odom(2);
                    for (unsigned int i=0; i<9; i++)
                      this->collision_msg_.front_covariance[i] = icp_odom_cov(i);
                  }
                  else
                  {
                    this->collision_msg_.front_odom_diff.x = 0.0;
                    this->collision_msg_.front_odom_diff.y = 0.0;
                    this->collision_msg_.front_odom_diff.theta = 0.0;
                    for (unsigned int i=0; i<9; i++)
                      this->collision_msg_.front_covariance[i] = 0.0;
                    icp_odom_calculated = false;
                    icp_is_collision = false;
                  }
                }
                else
                {
                  this->collision_msg_.front_odom_diff.x = 0.0;
                  this->collision_msg_.front_odom_diff.y = 0.0;
                  this->collision_msg_.front_odom_diff.theta = 0.0;
                  for (unsigned int i=0; i<9; i++)
                    this->collision_msg_.front_covariance[i] = 0.0;
                  icp_odom_calculated = false;
                  icp_is_collision = false;
                }
              }
              else
              {
                icp_succedded = false;
                icp_odom_calculated = false;
                icp_is_collision = false;
                icp_blink_pose_calculated = false;
                this->collision_msg_.front_icp_points = 0;
                ROS_ERROR("CollisionManagerAlgNode::mainNodeThread -> Failed to call service on topic %s",this->icp_client_.getService().c_str());
              }
              this->collision_msg_.front_icp_check = icp_is_collision;
              if (this->config_.debug)
              {
                this->debug_front_icp_last_scan_publisher_.publish(this->debug_front_icp_last_scan_msg_);
                this->debug_front_icp_ref_scan_publisher_.publish(this->debug_front_icp_ref_scan_msg_);
                if (icp_succedded)
                {
                  this->debug_front_after_icp_scan_msg_.header.frame_id = "new_front_laser";
                  this->debug_front_after_icp_scan_publisher_.publish(this->debug_front_after_icp_scan_msg_);

                  if (icp_blink_pose_calculated)
                    this->debug_front_icp_pose_publisher_.publish(this->debug_front_icp_pose_msg_);
                }
              }
            }
            if (this->is_rear_icp_input_data_ok_ && set_last_ranges(false))
            {
              if (icp_client_.call(rear_icp_srv_))
              {
                // ROS_INFO("CollisionManagerAlgNode::mainNodeThread -> Rear icp_client_ received a response from service server");
                // ROS_INFO_STREAM("   valid " << (int)rear_icp_srv_.response.valid);
                // ROS_INFO_STREAM("   new_laser_pose " << rear_icp_srv_.response.new_laser_pose);
                // ROS_INFO_STREAM("   covariance " << rear_icp_srv_.response.covariance[0] << ", " << rear_icp_srv_.response.covariance[1] << ", " << rear_icp_srv_.response.covariance[2] << ", " << rear_icp_srv_.response.covariance[3] << ", " << rear_icp_srv_.response.covariance[4] << ", " << rear_icp_srv_.response.covariance[5] << ", " << rear_icp_srv_.response.covariance[6] << ", " << rear_icp_srv_.response.covariance[7] << ", " << rear_icp_srv_.response.covariance[8]);
                // ROS_INFO_STREAM("   nvalid " << rear_icp_srv_.response.nvalid << "; min " << this->config_.icp_min_points);
                // ROS_INFO_STREAM("   error " << rear_icp_srv_.response.error);
                // ROS_INFO_STREAM("   error/points " << rear_icp_srv_.response.error/rear_icp_srv_.response.nvalid << "; max " << this->config_.icp_max_err_by_points);

                icp_succedded = true;
                //update collison msg data.
                this->collision_msg_.rear_icp_err = rear_icp_srv_.response.error/rear_icp_srv_.response.nvalid;
                this->collision_msg_.rear_icp_points = rear_icp_srv_.response.nvalid;
                // this->collision_msg_.rear_laser_disp = rear_icp_srv_.response.new_laser_pose;
                // for (unsigned int i=0; i<9; i++)
                //   this->collision_msg_.rear_covariance[i] = rear_icp_srv_.response.covariance[i];

                //calculate new base_link pose in fixed frame coordenates
                new_laser_pose << rear_icp_srv_.response.new_laser_pose.x, rear_icp_srv_.response.new_laser_pose.y, rear_icp_srv_.response.new_laser_pose.theta;
                covariances << rear_icp_srv_.response.covariance[0], rear_icp_srv_.response.covariance[1], rear_icp_srv_.response.covariance[2],
                                rear_icp_srv_.response.covariance[3], rear_icp_srv_.response.covariance[4], rear_icp_srv_.response.covariance[5],
                                rear_icp_srv_.response.covariance[6], rear_icp_srv_.response.covariance[7], rear_icp_srv_.response.covariance[8];
                
                icp_blink_pose_calculated = get_base_link_new_pose_from_icp(new_laser_pose, covariances, blink_pose, blink_cov, false);

                //publish new laser tf to visualize the laser points.
                if (this->config_.debug)
                {
                  this->transform_msg.header.frame_id = this->rear_laser_frame_;
                  this->transform_msg.child_frame_id = "new_rear_laser";
                  this->transform_msg.header.stamp = this->imu_stamp_;
                  this->transform_msg.transform.translation.x = rear_icp_srv_.response.new_laser_pose.x;
                  this->transform_msg.transform.translation.y = rear_icp_srv_.response.new_laser_pose.y;
                  this->transform_msg.transform.translation.z = 0.0;
                  tf2::Quaternion q;
                  q.setRPY(0.0, 0.0, rear_icp_srv_.response.new_laser_pose.theta);
                  this->transform_msg.transform.rotation = tf2::toMsg(q);
                  // this->transform_msg.transform.rotation.x = q.getRotation().x();
                  // this->transform_msg.transform.rotation.y = q.getRotation().y();
                  // this->transform_msg.transform.rotation.z = q.getRotation().z();
                  // this->transform_msg.transform.rotation.w = q.getRotation().w();
                  this->tf2_broadcaster.sendTransform(this->transform_msg);
                }

                if (rear_icp_srv_.response.nvalid > this->config_.icp_min_points && rear_icp_srv_.response.error/rear_icp_srv_.response.nvalid < this->config_.icp_max_err_by_points)
                {
                  if (get_icp_odometry(new_laser_pose, covariances, icp_odometry, icp_odom_cov, icp_odometry_diff, true))
                  {
                    icp_odom_calculated = true;
                    ROS_INFO_STREAM("Odom before collision " << this->odom_before_collision_);
                    ROS_INFO_STREAM("Odom after collision " << this->odom_after_collision_);
                    ROS_INFO_STREAM("New laser pose " << new_laser_pose);
                    ROS_INFO_STREAM("ICP odom " << icp_odometry);
                    ROS_INFO_STREAM("ICP odom diff " << icp_odometry_diff);
                    icp_is_collision = check_odometry_diff(icp_odometry_diff, diff_wheel_icp_odom);

                    // Update collision msg
                    this->collision_msg_.rear_odom_diff.x = diff_wheel_icp_odom(0);
                    this->collision_msg_.rear_odom_diff.y = diff_wheel_icp_odom(1);
                    this->collision_msg_.rear_odom_diff.theta = diff_wheel_icp_odom(2);
                    for (unsigned int i=0; i<9; i++)
                      this->collision_msg_.rear_covariance[i] = icp_odom_cov(i);
                  }
                  else
                  {
                    this->collision_msg_.rear_odom_diff.x = 0.0;
                    this->collision_msg_.rear_odom_diff.y = 0.0;
                    this->collision_msg_.rear_odom_diff.theta = 0.0;
                    for (unsigned int i=0; i<9; i++)
                      this->collision_msg_.rear_covariance[i] = 0.0;
                    icp_odom_calculated = false;
                    icp_is_collision = false;
                  }
                }
                else
                {
                  this->collision_msg_.rear_odom_diff.x = 0.0;
                  this->collision_msg_.rear_odom_diff.y = 0.0;
                  this->collision_msg_.rear_odom_diff.theta = 0.0;
                  for (unsigned int i=0; i<9; i++)
                    this->collision_msg_.rear_covariance[i] = 0.0;
                  icp_odom_calculated = false;
                  icp_is_collision = false;
                }
              }
              else
              {
                icp_succedded = false;
                icp_odom_calculated = false;
                icp_is_collision = false;
                icp_blink_pose_calculated = false;
                this->collision_msg_.rear_icp_points = 0;
                ROS_ERROR("CollisionManagerAlgNode::mainNodeThread -> Failed to call service on topic %s",this->icp_client_.getService().c_str());
              }
              this->collision_msg_.rear_icp_check = icp_is_collision;
              if (this->config_.debug)
              {
                this->debug_rear_icp_last_scan_publisher_.publish(this->debug_rear_icp_last_scan_msg_);
                this->debug_rear_icp_ref_scan_publisher_.publish(this->debug_rear_icp_ref_scan_msg_);
                if (icp_succedded)
                {
                  this->debug_rear_after_icp_scan_msg_.header.frame_id = "new_rear_laser";
                  this->debug_rear_after_icp_scan_publisher_.publish(this->debug_rear_after_icp_scan_msg_);

                  if (icp_blink_pose_calculated)
                    this->debug_rear_icp_pose_publisher_.publish(this->debug_rear_icp_pose_msg_);
                }
              }
            }
            // fuse poses?? call service to modify SLAM graph
          }
          ROS_INFO_STREAM("");
          ROS_INFO_STREAM("");
          ROS_INFO_STREAM("");
          //restart variables
          this->in_collision_ = false;
          this->low_acc_count_ = 0;
          this->low_ang_vel_count_ = 0;
          this->is_front_icp_input_data_ok_ = false;
          this->is_rear_icp_input_data_ok_ = false;

          //publish collision.
          this->collisions_publisher_.publish(this->collision_msg_);
        }
        else 
        {
          //Update max_acc
          if (acc_norm_2d > max_acc_norm_2d)
            this->max_acc_ = blink_acc;

          //Update angle rotated
          this->front_icp_srv_.request.rot_angle_estimation += blink_ang_vel(2)*(this->imu_stamp_ - this->last_imu_t_).toSec();
          this->last_imu_t_ = this->imu_stamp_;

          //Check collision timeout
          if ((ros::Time::now() - this->collision_start_stamp_) > ros::Duration(this->config_.collision_timeout))
          {
            ROS_WARN_STREAM("CollisionManagerAlgNode::mainNodeThread -> No end of collision detected after " << this->config_.collision_timeout << " seconds.");
            this->in_collision_ = false;
            this->low_acc_count_ = 0;
            this->low_ang_vel_count_ = 0;
          }
        }
      }
      else if (check_start_of_collision(acc_norm_2d, std::fabs(blink_ang_vel(2))))
      {
        ROS_INFO_STREAM("CollisionManagerAlgNode::mainNodeThread -> collision start detected.");
        this->collision_start_stamp_ = this->imu_stamp_;

        // Get reference scan
        if (!(this->config_.front_laser_en || this->config_.rear_laser_en))
          ROS_WARN("CollisionManagerAlgNode::mainNodeThread -> Collision detected but no laser is enabled.");
        else
        {
          this->is_front_icp_input_data_ok_ = set_ref_ranges(true);
          this->is_rear_icp_input_data_ok_ = set_ref_ranges(false);

          if (!this->is_front_icp_input_data_ok_)
            this->collision_msg_.front_icp_points = 0;
          if (!this->is_rear_icp_input_data_ok_)
            this->collision_msg_.rear_icp_points = 0;
        }

        this->odom_input_is_ok_ = set_odom_before_collision();

        //init variables
        this->in_collision_ = true;
        this->max_acc_ = blink_acc;
        this->last_imu_t_ = this->collision_start_stamp_;
        this->front_icp_srv_.request.rot_angle_estimation = 0.0;
        this->rear_icp_srv_.request.rot_angle_estimation = 0.0;

        //update collision msg data
        this->collision_msg_.header.stamp = this->imu_stamp_;
        this->collision_msg_.ang_vel_diff = std::fabs(blink_ang_vel(2));
      }
    }
    this->new_imu_input_ = false;
  }

  // [fill msg structures]
  // Initialize the topic message structure
  //this->collisions_collision_msg_.data = my_var;

  // Initialize the topic message structure
  //this->debug_rear_after_icp_scan_LaserScan_msg_.data = my_var;

  // Initialize the topic message structure
  //this->debug_front_after_icp_scan_LaserScan_msg_.data = my_var;

  // Initialize the topic message structure
  //this->debug_rear_icp_pose_PoseWithCovariance_msg_.data = my_var;

  // Initialize the topic message structure
  //this->debug_front_icp_pose_PoseWithCovariance_msg_.data = my_var;

  // Initialize the topic message structure
  //this->debug_rear_icp_last_scan_LaserScan_msg_.data = my_var;

  // Initialize the topic message structure
  //this->debug_rear_icp_ref_scan_LaserScan_msg_.data = my_var;

  // Initialize the topic message structure
  //this->debug_front_icp_last_scan_LaserScan_msg_.data = my_var;

  // Initialize the topic message structure
  //this->debug_front_icp_ref_scan_LaserScan_msg_.data = my_var;

  
  /*
  //tf2_listener example BEGIN
  try{
    std::string target_frame             = "child_frame";
    std::string source_frame             = "parent_frame";
    ros::Time time                       = ros::Time::now();
    ros::Duration timeout                = ros::Duration(0.1);
    this->alg_.unlock();
    bool tf2_exists = this->tf2_buffer.canTransform(target_frame, source_frame, time, timeout);
    this->alg_.lock();
    if(tf2_exists){
      geometry_msgs::PoseStamped stamped_pose_in;
      stamped_pose_in.header.stamp     = time;
      stamped_pose_in.header.frame_id  = source_frame;
      stamped_pose_in.pose.position.x    = 1.0;
      stamped_pose_in.pose.position.y    = 0.0;
      stamped_pose_in.pose.position.z    = 0.0;
      tf2::Quaternion quat_tf;
      quat_tf.setRPY(0.0, 0.0, 1.5709);
      geometry_msgs::Quaternion quat_msg;
      tf2::convert(quat_tf, quat_msg);
      stamped_pose_in.pose.orientation = quat_msg;
      double yaw, pitch, roll;
      tf2::Matrix3x3(tf2::Quaternion(quat_msg.x, quat_msg.y,quat_msg.z, quat_msg.w)).getEulerYPR(yaw, pitch, roll);
      ROS_INFO("Original    pose in '%s' frame, with (x,y,z)=[%f,%f,%f], (yaw,pitch,roll)=[%f,%f,%f]", stamped_pose_in.header.frame_id.c_str(), stamped_pose_in.pose.position.x, stamped_pose_in.pose.position.y, stamped_pose_in.pose.position.z, yaw, pitch, roll);
      geometry_msgs::PoseStamped stamped_pose_out;
      geometry_msgs::TransformStamped transform;
      transform = this->tf2_buffer.lookupTransform(target_frame, source_frame, time);
      quat_msg = transform.transform.rotation;
      tf2::Matrix3x3(tf2::Quaternion(quat_msg.x, quat_msg.y,quat_msg.z, quat_msg.w)).getEulerYPR(yaw, pitch, roll);
      ROS_INFO("Found transform betwen frames (%s-->%s) with (x,y,z)=[%f,%f,%f], (yaw,pitch,roll)=[%f,%f,%f]", source_frame.c_str(), target_frame.c_str(), transform.transform.translation.x, transform.transform.translation.y, transform.transform.translation.z, yaw, pitch, roll);
      tf2::doTransform(stamped_pose_in, stamped_pose_out, transform);
      quat_msg = stamped_pose_out.pose.orientation;
      tf2::Matrix3x3(tf2::Quaternion(quat_msg.x, quat_msg.y,quat_msg.z, quat_msg.w)).getEulerYPR(yaw, pitch, roll);
      ROS_INFO("Transformed pose in '%s' frame, with (x,y,z)=[%f,%f,%f], (yaw,pitch,roll)=[%f,%f,%f]", stamped_pose_out.header.frame_id.c_str(), stamped_pose_out.pose.position.x, stamped_pose_out.pose.position.y, stamped_pose_out.pose.position.z, yaw, pitch, roll);
      ROS_INFO("---");
    }else{
      ROS_WARN("No transform found from '%s' to '%s'", source_frame.c_str(), target_frame.c_str()); }
  }catch (tf2::TransformException &ex){
    ROS_ERROR("TF2 Exception: %s",ex.what()); }
  //tf2_listener example END
  */

  
  // [fill srv structure and make request to the server]
  //icp_srv_.request.data = my_var;
  //ROS_INFO("CollisionManagerAlgNode:: Calling service icp_client_!");
  //if (icp_client_.call(icp_srv_))
  //{
  //  ROS_INFO("CollisionManagerAlgNode:: icp_client_ received a response from service server");
  ////  ROS_INFO("CollisionManagerAlgNode:: Response: %s", icp_srv_.response.somestringfield.c_str());
  //}
  //else
  //{
  //  ROS_INFO("CollisionManagerAlgNode:: Failed to call service on topic %s",this->icp_client_.getService().c_str());
  //}


  
  // [fill action structure and make request to the action server]

  // [publish messages]
  // Uncomment the following line to publish the topic message
  //this->collisions_publisher_.publish(this->collisions_collision_msg_);

  // Uncomment the following line to publish the topic message
  //this->debug_rear_after_icp_scan_publisher_.publish(this->debug_rear_after_icp_scan_LaserScan_msg_);

  // Uncomment the following line to publish the topic message
  //this->debug_front_after_icp_scan_publisher_.publish(this->debug_front_after_icp_scan_LaserScan_msg_);

  
  /*
  //tf2_broadcaster example BEGIN
  this->transform_msg.header.stamp    = ros::Time::now();
  this->transform_msg.header.frame_id = "parent_frame";
  this->transform_msg.child_frame_id  = "child_frame";
  geometry_msgs::Transform t;
  t.translation.x = 0.0;
  t.translation.y = 0.0;
  t.translation.z = 0.0;
  tf2::Quaternion q;
  q.setRPY(0.0, 0.0, 0.0);
  t.rotation = tf2::toMsg(q);
  this->transform_msg.transform = t;
  this->tf2_broadcaster.sendTransform(this->transform_msg);
  //tf2_broadcaster example END
  */

  // Uncomment the following line to publish the topic message
  //this->debug_rear_icp_pose_publisher_.publish(this->debug_rear_icp_pose_PoseWithCovariance_msg_);

  // Uncomment the following line to publish the topic message
  //this->debug_front_icp_pose_publisher_.publish(this->debug_front_icp_pose_PoseWithCovariance_msg_);

  // Uncomment the following line to publish the topic message
  //this->debug_rear_icp_last_scan_publisher_.publish(this->debug_rear_icp_last_scan_LaserScan_msg_);

  // Uncomment the following line to publish the topic message
  //this->debug_rear_icp_ref_scan_publisher_.publish(this->debug_rear_icp_ref_scan_LaserScan_msg_);

  // Uncomment the following line to publish the topic message
  //this->debug_front_icp_last_scan_publisher_.publish(this->debug_front_icp_last_scan_LaserScan_msg_);

  // Uncomment the following line to publish the topic message
  //this->debug_front_icp_ref_scan_publisher_.publish(this->debug_front_icp_ref_scan_LaserScan_msg_);


  this->odom_mutex_exit();
  this->front_laser_scan_mutex_exit();
  this->rear_laser_scan_mutex_exit();
  this->imu_mutex_exit();
  // this->alg_.unlock();
}

bool CollisionManagerAlgNode::get_base_link_imu_data(Eigen::Vector3d& _global_acc, Eigen::Vector3d& _global_ang_vel)
{
  if (this->config_.base_link_frame != this->imu_frame_ || !this->tf_base_link_imu_loaded_)
  {
    geometry_msgs::TransformStamped tf_base_link_imu_msg;
    if (!this->tf2_buffer.canTransform(this->config_.base_link_frame, this->imu_frame_, ros::Time::now(), ros::Duration(this->config_.tf_timeout)))
    {
      ROS_WARN_STREAM("CollisionManagerAlgNode::get_base_link_imu_data -> Can't transform from " << this->config_.base_link_frame << " to " << this->imu_frame_ << " at " << ros::Time::now());
      this->tf_base_link_imu_loaded_ = false;
      return false;
    }
    tf_base_link_imu_msg = this->tf2_buffer.lookupTransform(this->config_.base_link_frame, this->imu_frame_, ros::Time::now());
    Eigen::Quaternion<double> rot(
    tf_base_link_imu_msg.transform.rotation.w, tf_base_link_imu_msg.transform.rotation.x, tf_base_link_imu_msg.transform.rotation.y, tf_base_link_imu_msg.transform.rotation.z);
    
    this->tf_base_link_imu_ = Eigen::Transform<double,3,Eigen::Affine>(rot);
    this->tf_base_link_imu_loaded_ = true;
  }
  _global_acc = this->tf_base_link_imu_*this->imu_local_acc_;
  _global_ang_vel = this->tf_base_link_imu_*this->imu_local_ang_vel_;
  return true;
}

bool CollisionManagerAlgNode::check_acc_collision_transition(double _acc_norm_2d, bool _start)
{
  if (!this->config_.collision_acc_transition_counter_en)
    return true;
  if ((_start && (_acc_norm_2d > this->config_.collision_acc_th)) || (!_start && (_acc_norm_2d < this->config_.collision_acc_th)))
    this->low_acc_count_++;
  else
  {
    this->low_acc_count_ = 0;
    return false;
  }
  return ((_start && (_acc_norm_2d > this->config_.collision_acc_th)) || (!_start && (_acc_norm_2d < this->config_.collision_acc_th)))  
         && (_start || (!_start && (!this->config_.collision_acc_transition_counter_en || (this->low_acc_count_ >= this->config_.collision_acc_counter_limit))));
}

bool CollisionManagerAlgNode::check_ang_vel_collision_transition(double _ang_vel, bool _start)
{
  double diff = std::fabs(_ang_vel - this->angular_vel_);
  if (!this->config_.collision_ang_vel_transition_counter_en)
    return true;
  if ((_start && (diff > this->config_.collision_ang_vel_th)) || (!_start && (diff < this->config_.collision_ang_vel_th)))
    this->low_ang_vel_count_++;
  else
  {
    this->low_ang_vel_count_ = 0;
    return false;
  }
  return ((_start && (diff > this->config_.collision_ang_vel_th)) || (!_start && (diff < this->config_.collision_ang_vel_th)))  
         && (_start || (!_start && (!this->config_.collision_ang_vel_transition_counter_en || (this->low_ang_vel_count_ >= this->config_.collision_ang_vel_counter_limit))));
}

bool CollisionManagerAlgNode::check_end_of_collision(double _acc_norm_2d, double _ang_vel)
{
  bool acc_true = check_acc_collision_transition(_acc_norm_2d, false);
  bool ang_vel_true = check_ang_vel_collision_transition(_ang_vel, false);
  // if (acc_true)
  //   ROS_INFO_STREAM("CollisionManagerAlgNode::check_end_of_collision -> (END) Acc true. acc: " << _acc_norm_2d << "; vel: " << _ang_vel);
  // if (ang_vel_true)
  //   ROS_INFO_STREAM("CollisionManagerAlgNode::check_end_of_collision -> (END) Angular vel true. acc: " << _acc_norm_2d << "; vel: " << _ang_vel);
  return acc_true && ang_vel_true;
}

bool CollisionManagerAlgNode::check_start_of_collision(double _acc_norm_2d, double _ang_vel)
{
  bool acc_true = check_acc_collision_transition(_acc_norm_2d, true);
  bool ang_vel_true = check_ang_vel_collision_transition(_ang_vel, true);

  // update collision msg
  if (acc_true)
    this->collision_msg_.acc_trigger = true;
  if (ang_vel_true)
    this->collision_msg_.ang_vel_trigger = true;
  // if (acc_true)
  //   ROS_INFO_STREAM("CollisionManagerAlgNode::check_start_of_collision -> (START) Acc true. acc: " << _acc_norm_2d << "; vel: " << _ang_vel);
  // if (ang_vel_true)
  //   ROS_INFO_STREAM("CollisionManagerAlgNode::check_start_of_collision -> (START) Angular vel true. acc: " << _acc_norm_2d << "; vel: " << _ang_vel);
  return acc_true || ang_vel_true;
}

bool CollisionManagerAlgNode::set_ref_ranges(bool _front)
{
  if (!(_front? this->config_.front_laser_en: this->config_.rear_laser_en))
    return false;
  auto it = (_front? this->front_ranges_.begin(): this->rear_ranges_.begin());
  while ((it != (_front? this->front_ranges_.end(): this->rear_ranges_.end())) && (it->stamp > (this->collision_start_stamp_ - ros::Duration(this->config_.scan_diff_t_from_collision))))
    ++it;
  if (it == (_front? this->front_ranges_.end(): this->rear_ranges_.end()))
  {
    ROS_WARN_STREAM("CollisionManagerAlgNode::set_ref_ranges -> Collision detected at " << this->collision_start_stamp_ << " but there is no " 
                      << (_front? "front ": "rear ") << "scan before collision.");
    if ((_front? this->front_ranges_: this->rear_ranges_).size() > 0)
      ROS_WARN_STREAM("CollisionManagerAlgNode::set_ref_ranges -> scan times: " << (_front? this->front_ranges_: this->rear_ranges_).front().stamp << ", " << (_front? this->front_ranges_: this->rear_ranges_).back().stamp);
    else
      ROS_WARN_STREAM("CollisionManagerAlgNode::set_ref_ranges -> No scan buffered");
    return false;
  }
  (_front? this->front_icp_srv_: this->rear_icp_srv_).request.ref_ranges = it->ranges;
  if (this->config_.debug)
  {
    (_front? this->debug_front_icp_ref_scan_msg_: this->debug_rear_icp_ref_scan_msg_).ranges = it->ranges;
    (_front? this->debug_front_icp_ref_scan_msg_: this->debug_rear_icp_ref_scan_msg_).header.stamp = it->stamp;
  }
  return true;
}

bool CollisionManagerAlgNode::set_odom_before_collision(void)
{
  auto it = this->odom_buffer_.begin();
  while ((it != this->odom_buffer_.end()) && (it->stamp > (this->collision_start_stamp_ - ros::Duration(this->config_.scan_diff_t_from_collision))))
    ++it;
  if (it == this->odom_buffer_.end())
  {
    ROS_WARN_STREAM("CollisionManagerAlgNode::set_odom_before_collision -> Collision detected at " << this->collision_start_stamp_ << " but there is no odometry before collision"); 
    if (this->odom_buffer_.size() > 0)
      ROS_WARN_STREAM("CollisionManagerAlgNode::set_odom_before_collision -> Odometry times: " << this->odom_buffer_.front().stamp << ", " << this->odom_buffer_.back().stamp);
    else
      ROS_WARN_STREAM("CollisionManagerAlgNode::set_odom_before_collision -> No Odometry buffered");
    return false;
  }
  this->odom_before_collision_ = it->odom;
  return true;
}

bool CollisionManagerAlgNode::set_last_ranges(bool _front)
{
  if (!(_front? this->config_.front_laser_en: this->config_.rear_laser_en))
    return false;

  if ((_front? this->front_ranges_: this->rear_ranges_).size() > 0)
  {
    (_front? this->front_icp_srv_: this->rear_icp_srv_).request.last_ranges = (_front? this->front_ranges_: this->rear_ranges_).front().ranges;
    if (this->config_.debug)
    {
      (_front? this->debug_front_icp_last_scan_msg_: this->debug_rear_icp_last_scan_msg_).ranges = (_front? this->front_ranges_: this->rear_ranges_).front().ranges;
      (_front? this->debug_front_icp_last_scan_msg_: this->debug_rear_icp_last_scan_msg_).header.stamp = (_front? this->front_ranges_: this->rear_ranges_).front().stamp;
      (_front? this->debug_front_after_icp_scan_msg_: this->debug_rear_after_icp_scan_msg_).ranges = (_front? this->front_ranges_: this->rear_ranges_).front().ranges;
      (_front? this->debug_front_after_icp_scan_msg_: this->debug_rear_after_icp_scan_msg_).header.stamp = (_front? this->front_ranges_: this->rear_ranges_).front().stamp;
    }
    return true;
  }
  ROS_WARN_STREAM("CollisionManagerAlgNode::set_last_ranges -> No scan buffered");
  return false;
}

bool CollisionManagerAlgNode::get_last_odom(void)
{
  if (this->odom_buffer_.size() > 0)
  {
    this->odom_after_collision_ = this->odom_buffer_.front().odom;
    return true;
  }
  ROS_WARN_STREAM("CollisionManagerAlgNode::get_last_odom -> No odom buffered");
  return false;
}

bool CollisionManagerAlgNode::get_base_link_new_pose_from_icp(const Eigen::Vector3d& _new_laser_pose, const Eigen::Matrix3d& _covariances, Eigen::Vector3d& _blink_pose, Eigen::Matrix3d& _blink_cov, const bool _front)
{
  geometry_msgs::TransformStamped tf_aux;
  tf2::Transform tf_map_laser, tf_map_new_laser, tf_laser_new_laser, tf_map_new_blink, tf_laser_map;

  if (!load_tf_blink_laser(_front))
    return false;

  if (!this->tf2_buffer.canTransform(this->config_.fixed_frame, (_front? this->front_laser_frame_: this->rear_laser_frame_), this->imu_stamp_, ros::Duration(this->config_.tf_timeout)))
  {
    ROS_WARN_STREAM("CollisionManagerAlgNode::get_base_link_new_pose_from_icp -> Can't transform from " << this->config_.fixed_frame << " to " << (_front? this->front_laser_frame_: this->rear_laser_frame_) << " at " << this->imu_stamp_);
    return false;
  }
  else
  {
    tf_aux = this->tf2_buffer.lookupTransform(this->config_.fixed_frame, (_front? this->front_laser_frame_: this->rear_laser_frame_), this->imu_stamp_);
    tf2::fromMsg(tf_aux.transform, tf_map_laser);
  }

  tf2::Quaternion q;
  q.setRPY(0.0, 0.0, _new_laser_pose(2));
  tf_laser_new_laser = tf2::Transform(q, tf2::Vector3(_new_laser_pose(0), _new_laser_pose(1), 0.0));
  tf_map_new_laser = tf_map_laser*tf_laser_new_laser;

  tf_map_new_blink = tf_map_new_laser*(_front? this->tf_blink_front_laser: this->tf_blink_rear_laser).inverse();
  _blink_pose << tf_map_new_blink.getOrigin().x(), tf_map_new_blink.getOrigin().y(), tf2::getYaw(tf_map_new_blink.getRotation());

  tf_laser_map = tf_map_laser.inverse();
  Eigen::Quaternion<double> rot(tf_laser_map.getRotation().w(), tf_laser_map.getRotation().x(), tf_laser_map.getRotation().y(), tf_laser_map.getRotation().z());
  
  Eigen::Transform<double,3,Eigen::Affine> laser_map_rot(rot);
  _blink_cov = laser_map_rot*_covariances;

  for (unsigned int i=0; i<9; i++)
    _blink_cov(i) = std::fabs(_blink_cov(i))*(_covariances(i) < 0? -1: 1);

  if (this->config_.debug)
  {
    if (_front)
    {
      this->debug_front_icp_pose_msg_.header.stamp = this->imu_stamp_;
      tf2::toMsg(tf_map_new_blink, this->debug_front_icp_pose_msg_.pose.pose);
      this->debug_front_icp_pose_msg_.pose.covariance[0] = _blink_cov(0);
      this->debug_front_icp_pose_msg_.pose.covariance[1] = _blink_cov(1);
      this->debug_front_icp_pose_msg_.pose.covariance[5] = _blink_cov(2);
      this->debug_front_icp_pose_msg_.pose.covariance[6] = _blink_cov(3);
      this->debug_front_icp_pose_msg_.pose.covariance[7] = _blink_cov(4);
      this->debug_front_icp_pose_msg_.pose.covariance[11] = _blink_cov(5);
      this->debug_front_icp_pose_msg_.pose.covariance[30] = _blink_cov(6);
      this->debug_front_icp_pose_msg_.pose.covariance[31] = _blink_cov(7);
      this->debug_front_icp_pose_msg_.pose.covariance[35] = _blink_cov(8);
    }
    else
    {
      this->debug_rear_icp_pose_msg_.header.stamp = this->imu_stamp_;
      tf2::toMsg(tf_map_new_blink, this->debug_rear_icp_pose_msg_.pose.pose);
      this->debug_rear_icp_pose_msg_.pose.covariance[0] = _blink_cov(0);
      this->debug_rear_icp_pose_msg_.pose.covariance[1] = _blink_cov(1);
      this->debug_rear_icp_pose_msg_.pose.covariance[5] = _blink_cov(2);
      this->debug_rear_icp_pose_msg_.pose.covariance[6] = _blink_cov(3);
      this->debug_rear_icp_pose_msg_.pose.covariance[7] = _blink_cov(4);
      this->debug_rear_icp_pose_msg_.pose.covariance[11] = _blink_cov(5);
      this->debug_rear_icp_pose_msg_.pose.covariance[30] = _blink_cov(6);
      this->debug_rear_icp_pose_msg_.pose.covariance[31] = _blink_cov(7);
      this->debug_rear_icp_pose_msg_.pose.covariance[35] = _blink_cov(8);
    }

  }
  return true;
}

bool CollisionManagerAlgNode::get_icp_odometry(const Eigen::Vector3d& _new_laser_pose, const Eigen::Matrix3d& _covariances, Eigen::Vector3d& _odom, Eigen::Matrix3d& _odom_cov, Eigen::Vector3d& _odom_diff, const bool _front)
{
  geometry_msgs::TransformStamped tf_aux;
  tf2::Transform tf_odom_laser, tf_odom_new_laser, tf_laser_new_laser, tf_odom_new_blink, tf_laser_odom, tf_odom_new_laser_diff, rot_odom_laser;

  if (!load_tf_blink_laser(_front))
    return false;

  if (!this->tf2_buffer.canTransform(this->odom_frame_, (_front? this->front_laser_frame_: this->rear_laser_frame_), this->imu_stamp_, ros::Duration(this->config_.tf_timeout)))
  {
    ROS_WARN_STREAM("CollisionManagerAlgNode::get_icp_odometry -> Can't transform from " << this->odom_frame_ << " to " << (_front? this->front_laser_frame_: this->rear_laser_frame_) << " at " << this->imu_stamp_);
    return false;
  }
  else
  {
    tf_aux = this->tf2_buffer.lookupTransform(this->odom_frame_, (_front? this->front_laser_frame_: this->rear_laser_frame_), this->imu_stamp_);
    tf2::fromMsg(tf_aux.transform, tf_odom_laser);
  }

  tf2::Quaternion q;
  q.setRPY(0.0, 0.0, _new_laser_pose(2));
  tf_laser_new_laser = tf2::Transform(q, tf2::Vector3(_new_laser_pose(0), _new_laser_pose(1), 0.0));
  tf_odom_new_laser = tf_odom_laser*tf_laser_new_laser;

  tf_odom_new_blink = tf_odom_new_laser*(_front? this->tf_blink_front_laser: this->tf_blink_rear_laser).inverse();
  _odom << tf_odom_new_blink.getOrigin().x(), tf_odom_new_blink.getOrigin().y(), tf2::getYaw(tf_odom_new_blink.getRotation());

  tf_laser_odom = tf_odom_laser.inverse();
  Eigen::Quaternion<double> rot(tf_laser_odom.getRotation().w(), tf_laser_odom.getRotation().x(), tf_laser_odom.getRotation().y(), tf_laser_odom.getRotation().z());
  
  Eigen::Transform<double,3,Eigen::Affine> laser_odom_rot(rot);
  _odom_cov = laser_odom_rot*_covariances;

  for (unsigned int i=0; i<9; i++)
    _odom_cov(i) = std::fabs(_odom_cov(i))*(_covariances(i) < 0? -1: 1);

  rot_odom_laser = tf2::Transform(tf_odom_laser.getRotation(), tf2::Vector3(0.0, 0.0, 0.0));
  tf_odom_new_laser_diff = rot_odom_laser*tf_laser_new_laser;
  _odom_diff <<  tf_odom_new_laser_diff.getOrigin().x(), tf_odom_new_laser_diff.getOrigin().y(), _new_laser_pose(2);

  return true;
}

bool CollisionManagerAlgNode::load_tf_blink_laser(bool _front)
{
  geometry_msgs::TransformStamped tf_aux;
  if ((_front? this->tf_base_link_front_laser_loaded_: this->tf_base_link_rear_laser_loaded_))
    return true;
  if (!this->tf2_buffer.canTransform(this->config_.base_link_frame, (_front? this->front_laser_frame_: this->rear_laser_frame_), this->imu_stamp_, ros::Duration(this->config_.tf_timeout)))
  {
    ROS_WARN_STREAM("CollisionManagerAlgNode::load_tf_blink_laser -> Can't transform from " << this->config_.base_link_frame << " to " << (_front? this->front_laser_frame_: this->rear_laser_frame_) << " at " << this->imu_stamp_);
    // this->tf_base_link_front_laser_loaded_ = false;
    return false;
  }
  else
  {
    tf_aux = this->tf2_buffer.lookupTransform(this->config_.base_link_frame, (_front? this->front_laser_frame_: this->rear_laser_frame_), this->imu_stamp_);
    tf2::fromMsg(tf_aux.transform, (_front? this->tf_blink_front_laser: this->tf_blink_rear_laser));
    (_front? this->tf_base_link_front_laser_loaded_: this->tf_base_link_rear_laser_loaded_) = true;
  }
  return true;
}

bool CollisionManagerAlgNode::check_odometry_diff(const Eigen::Vector3d& _icp_odometry, Eigen::Vector3d& _icp_odometry_diff)
{
  if (!this->odom_input_is_ok_)
  {
    ROS_WARN_STREAM("CollisionManagerAlgNode::check_odometry_diff -> No wheel odometry information has been saved. Odometry diff can't be calculated.");
    _icp_odometry_diff << 0.0, 0.0, 0.0;
    return false;
  }
  Eigen::Vector3d limits;
  limits << this->config_.odom_trans_th, this->config_.odom_trans_th, this->config_.odom_angular_th;
  _icp_odometry_diff = _icp_odometry - (this->odom_after_collision_ - this->odom_before_collision_);

  Eigen::Matrix<bool, 3, 1> odom_check = _icp_odometry.array().abs() >= limits.array();
  return (odom_check(0) || odom_check(1) || odom_check(2));
}

/*  [subscriber callbacks] */
void CollisionManagerAlgNode::odom_callback(const nav_msgs::Odometry::ConstPtr& msg)
{
  // ROS_INFO("CollisionManagerAlgNode::odom_callback: New Message Received");

  //use appropiate mutex to shared variables if necessary
  //this->alg_.lock();
  this->odom_mutex_enter();
  OdomWithStamp odom;
  tf2::Transform tf;
  tf2::fromMsg(msg->pose.pose, tf);
  odom.stamp = msg->header.stamp;
  odom.odom(0) = msg->pose.pose.position.x;
  odom.odom(1) = msg->pose.pose.position.y;
  odom.odom(2) = tf2::getYaw(tf.getRotation());
  if (this->odom_buffer_.size() >= this->config_.buffer_size)
    this->odom_buffer_.pop_back();
  this->odom_buffer_.push_front(odom);
  this->angular_vel_ = msg->twist.twist.angular.z;
  this->odom_frame_ = msg->header.frame_id;

  //std::cout << msg->data << std::endl;
  //unlock previously blocked shared variables
  //this->alg_.unlock();
  this->odom_mutex_exit();
}

void CollisionManagerAlgNode::odom_mutex_enter(void)
{
  pthread_mutex_lock(&this->odom_mutex_);
}

void CollisionManagerAlgNode::odom_mutex_exit(void)
{
  pthread_mutex_unlock(&this->odom_mutex_);
}

void CollisionManagerAlgNode::rear_laser_scan_callback(const sensor_msgs::LaserScan::ConstPtr& msg)
{
  // ROS_INFO("CollisionManagerAlgNode::rear_laser_scan_callback: New Message Received");

  //use appropiate mutex to shared variables if necessary
  //this->alg_.lock();
  this->rear_laser_scan_mutex_enter();
  if (this->config_.rear_laser_en)
  {
    this->rear_laser_frame_ = msg->header.frame_id;
    rear_icp_srv_.request.angle_min = msg->angle_min;
    rear_icp_srv_.request.angle_max = msg->angle_max;
    rear_icp_srv_.request.angle_increment = msg->angle_increment;
    rear_icp_srv_.request.scan_time = msg->scan_time;
    rear_icp_srv_.request.range_min = msg->range_min;
    rear_icp_srv_.request.range_max = msg->range_max;
    debug_rear_icp_ref_scan_msg_.angle_min = msg->angle_min;
    debug_rear_icp_ref_scan_msg_.angle_max = msg->angle_max;
    debug_rear_icp_ref_scan_msg_.angle_increment = msg->angle_increment;
    debug_rear_icp_ref_scan_msg_.scan_time = msg->scan_time;
    debug_rear_icp_ref_scan_msg_.range_min = msg->range_min;
    debug_rear_icp_ref_scan_msg_.range_max = msg->range_max;
    debug_rear_icp_ref_scan_msg_.header.frame_id = msg->header.frame_id;
    debug_rear_icp_last_scan_msg_.angle_min = msg->angle_min;
    debug_rear_icp_last_scan_msg_.angle_max = msg->angle_max;
    debug_rear_icp_last_scan_msg_.angle_increment = msg->angle_increment;
    debug_rear_icp_last_scan_msg_.scan_time = msg->scan_time;
    debug_rear_icp_last_scan_msg_.range_min = msg->range_min;
    debug_rear_icp_last_scan_msg_.range_max = msg->range_max;
    debug_rear_icp_last_scan_msg_.header.frame_id = msg->header.frame_id;
    debug_rear_after_icp_scan_msg_.angle_min = msg->angle_min;
    debug_rear_after_icp_scan_msg_.angle_max = msg->angle_max;
    debug_rear_after_icp_scan_msg_.angle_increment = msg->angle_increment;
    debug_rear_after_icp_scan_msg_.scan_time = msg->scan_time;
    debug_rear_after_icp_scan_msg_.range_min = msg->range_min;
    debug_rear_after_icp_scan_msg_.range_max = msg->range_max;

    RangesWithStamp r;
    r.ranges = msg->ranges;
    r.stamp = msg->header.stamp;
    if (this->rear_ranges_.size() >= this->config_.buffer_size)
      this->rear_ranges_.pop_back();
    this->rear_ranges_.push_front(r);
  }
  //std::cout << msg->data << std::endl;
  //unlock previously blocked shared variables
  //this->alg_.unlock();
  this->rear_laser_scan_mutex_exit();
}

void CollisionManagerAlgNode::rear_laser_scan_mutex_enter(void)
{
  pthread_mutex_lock(&this->rear_laser_scan_mutex_);
}

void CollisionManagerAlgNode::rear_laser_scan_mutex_exit(void)
{
  pthread_mutex_unlock(&this->rear_laser_scan_mutex_);
}

void CollisionManagerAlgNode::front_laser_scan_callback(const sensor_msgs::LaserScan::ConstPtr& msg)
{
  // ROS_INFO("CollisionManagerAlgNode::front_laser_scan_callback: New Message Received");

  //use appropiate mutex to shared variables if necessary
  //this->alg_.lock();
  this->front_laser_scan_mutex_enter();
  if (this->config_.front_laser_en)
  {
    this->front_laser_frame_ = msg->header.frame_id;
    front_icp_srv_.request.angle_min = msg->angle_min;
    front_icp_srv_.request.angle_max = msg->angle_max;
    front_icp_srv_.request.angle_increment = msg->angle_increment;
    front_icp_srv_.request.scan_time = msg->scan_time;
    front_icp_srv_.request.range_min = msg->range_min;
    front_icp_srv_.request.range_max = msg->range_max;
    debug_front_icp_ref_scan_msg_.angle_min = msg->angle_min;
    debug_front_icp_ref_scan_msg_.angle_max = msg->angle_max;
    debug_front_icp_ref_scan_msg_.angle_increment = msg->angle_increment;
    debug_front_icp_ref_scan_msg_.scan_time = msg->scan_time;
    debug_front_icp_ref_scan_msg_.range_min = msg->range_min;
    debug_front_icp_ref_scan_msg_.range_max = msg->range_max;
    debug_front_icp_ref_scan_msg_.header.frame_id = msg->header.frame_id;
    debug_front_icp_last_scan_msg_.angle_min = msg->angle_min;
    debug_front_icp_last_scan_msg_.angle_max = msg->angle_max;
    debug_front_icp_last_scan_msg_.angle_increment = msg->angle_increment;
    debug_front_icp_last_scan_msg_.scan_time = msg->scan_time;
    debug_front_icp_last_scan_msg_.range_min = msg->range_min;
    debug_front_icp_last_scan_msg_.range_max = msg->range_max;
    debug_front_icp_last_scan_msg_.header.frame_id = msg->header.frame_id;
    debug_front_after_icp_scan_msg_.angle_min = msg->angle_min;
    debug_front_after_icp_scan_msg_.angle_max = msg->angle_max;
    debug_front_after_icp_scan_msg_.angle_increment = msg->angle_increment;
    debug_front_after_icp_scan_msg_.scan_time = msg->scan_time;
    debug_front_after_icp_scan_msg_.range_min = msg->range_min;
    debug_front_after_icp_scan_msg_.range_max = msg->range_max;

    RangesWithStamp r;
    r.ranges = msg->ranges;
    r.stamp = msg->header.stamp;
    if (this->front_ranges_.size() >= this->config_.buffer_size)
      this->front_ranges_.pop_back();
    this->front_ranges_.push_front(r);
  }

  //std::cout << msg->data << std::endl;
  //unlock previously blocked shared variables
  //this->alg_.unlock();
  this->front_laser_scan_mutex_exit();
}

void CollisionManagerAlgNode::front_laser_scan_mutex_enter(void)
{
  pthread_mutex_lock(&this->front_laser_scan_mutex_);
}

void CollisionManagerAlgNode::front_laser_scan_mutex_exit(void)
{
  pthread_mutex_unlock(&this->front_laser_scan_mutex_);
}

void CollisionManagerAlgNode::imu_callback(const sensor_msgs::Imu::ConstPtr& msg)
{
  // ROS_INFO("CollisionManagerAlgNode::imu_callback: New Message Received");

  //use appropiate mutex to shared variables if necessary
  //this->alg_.lock();
  this->imu_mutex_enter();
  this->new_imu_input_ = true;
  this->imu_frame_ = msg->header.frame_id;
  this->imu_local_acc_ = Eigen::Vector3d(msg->linear_acceleration.x, msg->linear_acceleration.y, msg->linear_acceleration.z);
  this->imu_local_ang_vel_ = Eigen::Vector3d(msg->angular_velocity.x, msg->angular_velocity.y, msg->angular_velocity.z);
  this->imu_stamp_ = msg->header.stamp;
  // if (!this->in_collision_)
  //   this->collision_start_stamp_ = msg->header.stamp;
  // if (this->in_collision_)
  //   this->collision_end_stamp_ = msg->header.stamp;
  //std::cout << msg->data << std::endl;
  //unlock previously blocked shared variables
  //this->alg_.unlock();
  this->imu_mutex_exit();
}

void CollisionManagerAlgNode::imu_mutex_enter(void)
{
  pthread_mutex_lock(&this->imu_mutex_);
}

void CollisionManagerAlgNode::imu_mutex_exit(void)
{
  pthread_mutex_unlock(&this->imu_mutex_);
}


/*  [service callbacks] */

/*  [action callbacks] */

/*  [action requests] */

void CollisionManagerAlgNode::node_config_update(Config &config, uint32_t level)
{
  this->alg_.lock();
  this->imu_mutex_enter();
  this->rear_laser_scan_mutex_enter();
  this->front_laser_scan_mutex_enter();

  if(config.rate!=this->getRate())
    this->setRate(config.rate);
  this->config_=config;

  this->front_laser_scan_mutex_exit();
  this->rear_laser_scan_mutex_exit();
  this->imu_mutex_exit();
  this->alg_.unlock();
}

void CollisionManagerAlgNode::addNodeDiagnostics(void)
{
}

/* main function */
int main(int argc,char *argv[])
{
  return algorithm_base::main<CollisionManagerAlgNode>(argc, argv, "collision_manager_alg_node");
}
